"""Common docstrings."""

from textwrap import dedent

import pandas.util._decorators as pd_dec

from .options import OPTIONS


def docfiller(*args, **kwargs):
    """To fill common docs.

    Taken from pandas.utils._decorators

    """

    if OPTIONS["doc_sub"]:
        return pd_dec.doc(*args, **kwargs)
    else:

        def decorated(func):
            return func

        return decorated


# fmt: off
_shared_docs = {
    "copy":
    """
    copy : bool, optional
        If True, copy the data.  If False, attempt to use view.
    """,
    "copy_kws":
    """
    copy_kws : mapping, optional
        extra arguments to copy
    """,
    "lnz":
    """
    lnz: float or sequence of floats
        Value(s) of ``lnz`` (log of activity)
    """,
    "state_kws":
    """
    state_kws : Mapping, optiona
        'State' variables.  Common key/value pairs used by other methods:

        * beta (float) : inverse temperature
        * volume (float) : volume
    """,
    "extra_kws":
    """
    extra_kws : Mapping, Optional
        Extra parameters.  These are passed along to new objects without modification.
        Common key/value pairs:

    * PE (array-like) : value of total potential energy.
      Used in calculation of canonical and grand canonical average potential energy.
    """,
    "fill_value":
    """
    fill_value : scalar, default=nan
        Value to fill for masked elements.
    """,
    "data":
    """
    data : array-like
        Value of lnPi.  Number of dimensions should be same as ``len(lnz)``.
        For single component, ``data[k]`` is the value of lnPi for `k` particles.
    """,
    "mask_masked":
    """
    mask : bool or array-like, optional
        Where `mask` is `True`, values are "masked out".
    """,
    "mask_image":
    """
    mask : bool or array-like, optional
        This uses the "image" convention.  Where `mask` is `True`, values are include.
    """,
    "masks_masked":
    """
    masks : sequence of array-like of bools
        Value of `masks[i]` is the mask for ith lnPi.
        masks[i] is True where values are "masked out".
    """,
    "mask_convention":
    """
    convention : string or bool
        Convention for mask.  Allowable values are:

        * 'image' or True : `True` values included, `False` values excluded.
          This is the normal convention in :mod:`scipy.ndimage`.
        * 'masked' or False:  `False` values are included, `True` values are excluded.
          This is the convention in :mod:`numpy.ma`
    """,
    "labels":
    """
    labels : array-like of ints
        Where `labels == k`, indicates regions where data corresponds to the kth label.
    """,
    "base":
    """
    base : Data
        lnPi data object.
    """,
    "peak_kws":
    """
    peak_kws : mapping, optional
        Optional parameters to :func:`~lnpy.segment.peak_local_max_adaptive`
    """,

    "watershed_kws":
    """
    watershed_kws : mapping, optional
        Optional parameters to :func:`~skimage.segmentation.watershed`
    """,




}
# fmt: on


def _prepare_shared_docs(shared_docs):
    return {k: dedent(v).strip() for k, v in shared_docs.items()}


_shared_docs = _prepare_shared_docs(_shared_docs)


# add in xr_params
# _shared_docs["xr_params"] = "\n".join(
#     [
#         _shared_docs[k]
#         for k in ["dims", "mom_dims", "attrs", "coords", "name", "indexes", "template"]
#     ]
# )


docfiller_shared = docfiller(**_shared_docs)



"""
Alternative to ListIndex.  Will wrap stuff in Series
"""
from __future__ import absolute_import, division, print_function

import numpy as np
import pandas as pd
import xarray as xr

from .cached_decorators import gcached
from .extensions import AccessorMixin
from .utils import get_tqdm_build as get_tqdm
from .utils import labels_to_masks, masks_to_labels
from .utils import parallel_map_build as parallel_map


class SeriesWrapper(AccessorMixin):
    """
    wrap object in series
    """

    def __init__(
        self, data=None, index=None, dtype=None, name=None, base_class="first"
    ):

        if isinstance(data, self.__class__):
            x = data
            data = x.s

        self._base_class = base_class
        self._verify = self._base_class is not None

        series = pd.Series(data=data, index=index, dtype=dtype, name=name)
        self._verify_series(series)
        self._series = series
        self._cache = {}

    def _verify_series(self, series):
        if self._verify:
            base_class = self._base_class
            if isinstance(base_class, str) and base_class.lower() == "first":
                base_class = type(series.iloc[0])
            for d in series:
                if not issubclass(type(d), base_class):
                    raise ValueError(
                        "all elements must be of type {}".format(base_class)
                    )

    @property
    def series(self):
        """View of the underlying :class:`pandas.Series`"""
        return self._series

    @series.setter
    def series(self, series):
        self._cache = {}
        self._verify_series(series)
        self._series = series

    @property
    def s(self):
        """Alias to `self.series`"""
        return self.series

    def __iter__(self):
        return iter(self._series)

    def __next__(self):
        return next(self._series)

    @property
    def values(self):
        """Series values"""
        return self._series.values

    @property
    def items(self):
        """Alias to `self.values`"""
        return self.values

    @property
    def index(self):
        """Series index"""
        return self._series.index

    @property
    def name(self):
        """Series name"""
        return self._series.name

    def copy(self):
        return type(self)(data=self.s, base_class=self._base_class)

    def new_like(self, data=None, index=None, **kwargs):
        """Create new object with optional new data/index"""
        return self.__class__(
            data=data,
            index=index,
            dtype=self.s.dtype,
            name=self.s.name,
            base_class=self._base_class,
            **kwargs
        )

    def _wrapped_pandas_method(self, mtd, wrap=False, *args, **kwargs):
        """Wrap a generic pandas method to ensure it returns a GeoSeries"""
        val = getattr(self._series, mtd)(*args, **kwargs)
        if wrap and type(val) == pd.Series:
            val = self.new_like(val)
        return val

    def __getitem__(self, key):
        """Interface to :meth:`pandas.Series.__getitem__`"""
        return self._wrapped_pandas_method("__getitem__", wrap=True, key=key)

    def xs(self, key, axis=0, level=None, drop_level=False, wrap=True):
        """Interface to :meth:`pandas.Series.xs`"""
        return self._wrapped_pandas_method(
            "xs", wrap=wrap, key=key, axis=axis, level=level, drop_level=drop_level
        )

    def __setitem__(self, idx, values):
        """Interface to :meth:`pandas.Series.__setitem__`"""
        self._series[idx] = values

    def __repr__(self):
        return "<class {}>\n{}".format(self.__class__.__name__, repr(self.s))

    def __str__(self):
        return str(self.s)

    def __len__(self):
        return len(self.s)

    def append(
        self,
        to_append,
        ignore_index=False,
        verify_integrity=True,
        concat_kws=None,
        inplace=False,
    ):
        """Interface to :meth:`pandas.Series.append`


        Parameters
        ----------
        to_append : object
            Object to append
        ignore_index : bool, default=False
        verify_integrity : bool, default=True
        concat_kws : mapping, optional
            Extra arguments to
        inplace : bool, default = False

        See Also
        --------
        pandas.Series.append
        """
        if isinstance(to_append, self.__class__):
            to_append = to_append.series

        if concat_kws is None:
            concat_kws = {}

        s = pd.concat(
            (self.series, to_append),
            ignore_index=ignore_index,
            verify_integrity=verify_integrity,
            **concat_kws
        )

        # s = self._series.append(
        #     to_append, ignore_index=ignore_index, verify_integrity=verify_integrity
        # )

        if inplace:
            self.series = s
        else:
            return self.new_like(s)

    def droplevel(self, level, axis=0):
        """New object with dropped level

        See Also
        --------
        pandas.Series.droplevel
        """
        return self.new_like(self._series.droplevel(level=level, axis=axis))

    def apply(self, func, convert_dtype=True, args=(), wrap=False, **kwds):
        """Interface to :meth:`pandas.Series.apply`"""

        return self._wrapped_pandas_method(
            "apply",
            wrap=wrap,
            func=func,
            convert_dtype=convert_dtype,
            args=args,
            **kwds
        )

    def sort_index(self, wrap=True, *args, **kwargs):
        """Interface to :meth:`pandas.Series.sort_index`"""
        return self._wrapped_pandas_method("sort_index", wrap=wrap, *args, **kwargs)

    def groupby(
        self,
        by=None,
        axis=0,
        level=None,
        as_index=True,
        sort=True,
        group_keys=True,
        # squeeze=False,
        observed=False,
        wrap=False,
        **kwargs
    ):
        """
        Wrapper around :meth:`pandas.Series.groupby`.

        Paremters
        ---------
        wrap : bool, default=False
            if True, try to wrap output in class of self

        See Also
        --------
        pandas.Series.groupby
        """

        group = self.s.groupby(
            by=by,
            axis=axis,
            level=level,
            as_index=as_index,
            sort=sort,
            group_keys=group_keys,
            # squeeze=squeeze,
            observed=observed,
            **kwargs
        )
        if wrap:
            return _Groupby(self, group)
        else:
            return group

    def groupby_allbut(self, drop, **kwargs):
        """
        groupby all but columns in drop
        """
        from .utils import allbut

        if not isinstance(drop, list):
            drop = [drop]
        by = allbut(self.index.names, *drop)
        return self.groupby(by=by, **kwargs)

    @classmethod
    def _concat_to_series(cls, objs, **concat_kws):
        from collections.abc import Mapping, Sequence

        if isinstance(objs, Sequence):
            first = objs[0]
            if isinstance(first, cls):
                objs = (x._series for x in objs)
        elif isinstance(objs, Mapping):
            out = {}
            remap = None
            for k in objs:
                v = objs[k]
                if remap is None:
                    if isinstance(v, cls):
                        remap = True
                    else:
                        remap = False
                if remap:
                    out[k] = v._series
                else:
                    out[k] = v
            objs = out
        else:
            raise ValueError("bad input type {}".format(type(first)))
        return pd.concat(objs, **concat_kws)

    def concat_like(self, objs, **concat_kws):
        """Concat a sequence of objects like `self`"""
        s = self._concat_to_series(objs, **concat_kws)
        return self.new_like(s)

    @classmethod
    def concat(cls, objs, concat_kws=None, *args, **kwargs):
        """Create collection from sequence of objects"""
        if concat_kws is None:
            concat_kws = {}
        s = cls._concat_to_series(objs, **concat_kws)
        return cls(s, *args, **kwargs)


# Accessors
class _CallableResult(object):
    def __init__(self, parent, series):
        self._parent = parent
        self._series = series

    def __call__(self, *args, **kwargs):
        return self._parent.new_like(self._series(*args, **kwargs))


class _Groupby(object):
    def __init__(self, parent, group):
        self._parent = parent
        self._group = group

    def __iter__(self):
        return ((meta, self._parent.new_like(x)) for meta, x in self._group)

    def __getattr__(self, attr):
        if hasattr(self._group, attr):
            out = getattr(self._group, attr)
            if callable(out):
                return _CallableResult(self._parent, out)
            else:
                return self._parent.new_like(out)
        else:
            raise AttributeError("no attribute {} in groupby".format(attr))


@SeriesWrapper.decorate_accessor("loc")
class _LocIndexer(object):
    """
    Indexer by value.

    See :attr:`pandas.Series.loc`
    """

    def __init__(self, parent):
        self._parent = parent
        self._loc = self._parent._series.loc

    def __getitem__(self, idx):
        out = self._loc[idx]
        if isinstance(out, pd.Series):
            out = self._parent.new_like(out)
        return out

    def __setitem__(self, idx, values):
        self._parent._series.loc[idx] = values


@SeriesWrapper.decorate_accessor("iloc")
class _iLocIndexer(object):
    """Indexer by position.

    See :attr:`pandas.Series.iloc`
    """

    def __init__(self, parent):
        self._parent = parent
        self._iloc = self._parent._series.iloc

    def __getitem__(self, idx):
        out = self._iloc[idx]
        if isinstance(out, pd.Series):
            out = self._parent.new_like(out)
        return out

    def __setitem__(self, idx, values):
        self._parent._series.iloc[idx] = values


@SeriesWrapper.decorate_accessor("query")
class _Query(object):
    """
    Select values by string query.

    See :meth:`pandas.DataFrame.query`
    """

    def __init__(self, parent):
        self._parent = parent
        self._frame = self._parent.index.to_frame().reset_index(drop=True)

    def __call__(self, expr, **kwargs):
        idx = self._frame.query(expr, **kwargs).index
        return self._parent.iloc[idx]


class lnPiCollection(SeriesWrapper):
    # class lnPiCollection:
    r"""
    Wrapper around :class:`pandas.Series` for collection of :class:`lnpy.lnPiMasked` objects.


    Parameters
    ----------
    data : sequence of lnPiMasked
        :math:`\ln \Pi(N)` instances to consider.
    index : array-like, pandas.Index, pandas.MultiIndex, optional
        Index to apply to Series.
    xarray_output : bool, default = True
        If True, then wrap lnPiCollection outputs in :class:`~xarray.DataArray`
    concat_dim : str, optional
        Name of dimensions to concat results along.
        Also Used by :class:`~lnpy.ensembles.xGrandCanonical`.
    concat_coords : string, optional
        parameters `coords `to :func:`xarray.concat`
    unstack : bool, default=True
        If True, then outputs will be unstacked using :meth:`xarray.DataArray.unstack`
    *args **kwargs
        Extra arguments to Series constructor

    """

    _concat_dim = "sample"
    _concat_coords = "different"
    _use_joblib = True
    _xarray_output = True
    _xarray_unstack = True
    _xarray_dot_kws = {"optimize": "optimal"}
    _use_cache = True

    def __init__(
        self,
        data,
        index=None,
        xarray_output=True,
        concat_dim=None,
        concat_coords=None,
        unstack=True,
        *args,
        **kwargs
    ):

        if concat_dim is not None:
            self._concat_dim = concat_dim
        if concat_coords is not None:
            self._concat_coords = concat_coords
        if xarray_output is not None:
            self._xarray_output = xarray_output
        if unstack is not None:
            self._xarray_unstack = unstack

        super().__init__(data=data, index=index, *args, **kwargs)

        # update index name:
        # self._series.index.name = self._concat_dim

    def new_like(self, data=None, index=None):
        """Create new object with optional new data/index."""
        return super().new_like(
            data=data,
            index=index,
            concat_dim=self._concat_dim,
            concat_coords=self._concat_coords,
            xarray_output=self._xarray_output,
            unstack=self._xarray_unstack,
        )

    def _verify_series(self, series):
        super()._verify_series(series)
        if self._verify:
            first = series.iloc[0]
            state_kws = first.state_kws
            shape = first.shape
            # _base  = first._base

            for lnpi in series:
                assert lnpi.state_kws == state_kws
                assert lnpi.shape == shape
                # would like to do this, but
                # fails for parallel builds
                # assert lnpi._base is _base

    # repr
    @gcached()
    def _lnz_series(self):
        return self._series.apply(lambda x: x.lnz)

    def __repr__(self):
        return "<class {}>\n{}".format(self.__class__.__name__, repr(self._lnz_series))

    def __str__(self):
        return str(self._lnz_series)

    @property
    def state_kws(self):
        """state_kws from first :class:`~lnpy.lnPiMasked`"""
        return self.iloc[0].state_kws

    @property
    def nlnz(self):
        """Number of unique lnzs"""
        return len(self.index.droplevel("phase").drop_duplicates())

    @gcached()
    def index_frame(self):
        """
        DataFrame of values for each sample

        includes a column 'lnz_index' which is the unique lnz values
        regardless of phase
        """
        sample_frame = (
            self.index.droplevel("phase")
            .drop_duplicates()
            .to_frame()
            .assign(lnz_sample=lambda x: np.arange(len(x)))["lnz_sample"]
        )
        index_frame = (
            self.index.to_frame()
            .reset_index("phase", drop=True)[["phase"]]
            .assign(lnz_index=lambda x: sample_frame[x.index])
            .reset_index()
        )
        return index_frame

    def _get_lnz(self, component=None, iloc=0, zloc=None):
        """
        helper function to
        returns self.iloc[idx].lnz[component]
        """
        if zloc is not None:
            s = self.zloc[zloc]._series
        else:
            s = self._series
        lnz = s.iloc[iloc].lnz
        if component is not None:
            lnz = lnz[component]
        return lnz

    def _get_level(self, level="phase"):
        """
        return level values from index
        """
        index = self.index
        if isinstance(index, pd.MultiIndex):
            level_idx = index.names.index(level)
            index = index.levels[level_idx]
        return index

    def get_index_level(self, level="phase"):
        """Get index values for specified level"""
        return self._get_level(level=level)

    # @gcached()
    @property
    def _nrec(self):
        return len(self._series)

    def _lnpi_tot(self, fill_value=None):
        # old method
        # return np.stack([x.filled() for x in self])

        # new method
        # this is no faster than the original
        # but makes clear where the time is being spent
        first = self.iloc[0]
        n = len(self)
        out = np.empty((n,) + first.shape, dtype=first.dtype)
        seq = get_tqdm((x.filled(fill_value) for x in self), total=n)
        for i, x in enumerate(seq):
            out[i, ...] = x
        return out

    # @property
    # def _lnpi_0_tot(self):
    #     return np.array([x.data.ravel()[0] for x in self])

    @property
    def _lnz_tot(self):
        return np.stack([x.lnz for x in self])

    def _pi_params(self, fill_value=None):
        first = self.iloc[0]
        n = len(self)

        pi_norm = np.empty((n,) + first.shape, dtype=first.dtype)
        pi_sum = np.empty(n, dtype=first.dtype)
        lnpi_zero = np.empty(n, dtype=first.dtype)

        seq = get_tqdm(
            (x._pi_params(fill_value) for x in self), total=n, desc="pi_norm"
        )

        for i, (_pi_norm, _pi_sum, _lnpi_zero) in enumerate(seq):
            pi_norm[i, ...] = _pi_norm
            pi_sum[i, ...] = _pi_sum
            lnpi_zero[i, ...] = _lnpi_zero
        return pi_norm, pi_sum, lnpi_zero

    def wrap_list_results(self, items):
        """Unitility to wrap output in :class:xarray.DataArray"""
        if self._xarray_output:
            x = items[0]
            if isinstance(x, xr.DataArray):
                return xr.concat(items, self.index, coords=self._concat_coords)
            # else try to make array
        else:
            return items

    ##################################################
    # Constructors
    @classmethod
    def from_list(cls, items, index, *args, **kwargs):
        """
        Create collection from list of :class:`lnpy.lnPiMasked` objects.

        Parameters
        ----------
        items : sequence of lnPiMasked
            Sequence of lnPi
        index : sequence
            Sequence of phases ID for each lnPi
        *args
            Extra positional arguments to `cls`
        **kwargs :
            Extra keyword arguments to `cls`
        Returns
        -------
        output : class instance
        """
        df = pd.DataFrame(
            [lnpi._index_dict(phase) for lnpi, phase in zip(items, index)]
        )
        index = pd.MultiIndex.from_frame(df)
        return cls(data=items, index=index, *args, **kwargs)

    @classmethod
    def from_builder(
        cls,
        lnzs,
        build_phases,
        ref=None,
        build_kws=None,
        nmax=None,
        concat_kws=None,
        base_class="first",
        *args,
        **kwargs
    ):
        """
        Build collection from scalar builder

        Parameters
        ----------
        lnzs : 1-D sequence
            lnz value for the varying value
        ref : lnpi_phases object
            lnpi_phases to reweight to get list of lnpi's
        build_phases : callable
            Typically one of `PhaseCreator.build_phases_mu` or `PhaseCreator.build_phases_dmu`
        build_kws : optional
            optional arguments to `build_phases`
        Returns
        -------
        out : Collection object


        See Also
        --------
        ~lnpy.segment.Segmenter.build_phases
        ~lnpy.segment.Segmenter.build_phases_mu
        ~lnpy.segment.Segmenter.build_phases_dmu

        """
        if build_kws is None:
            build_kws = {}
        build_kws = dict(build_kws, phases_factory="None")
        seq = get_tqdm(lnzs, desc="build")
        L = parallel_map(build_phases, seq, ref=ref, nmax=nmax, **build_kws)
        # return cls.concat(L, verify=verify, concat_kws=concat_kws, base_class=base_class,
        #                   *args, **kwargs)

        items = []
        index = []
        for data, idx in L:
            items += data
            index += list(idx)
        return cls.from_list(items, index, base_class=base_class, *args, **kwargs)

    ################################################################################
    # dataarray io
    def to_dataarray(self, dtype=np.uint8, reset_index=True, **kwargs):
        """
        Convert collection to a :class:`~xarray.DataArray`
        """
        labels = []
        indexes = []

        for meta, g in self.groupby_allbut("phase"):
            indexes.append(g.index[[0]])

            features = np.array(g.index.get_level_values("phase")) + 1
            masks = [x.mask for x in g]
            labels.append(
                masks_to_labels(masks, features=features, convention=False, dtype=dtype)
            )

        index = indexes[0].append(indexes[1:])

        data = np.stack(labels)

        out = (
            xr.DataArray(
                data,
                dims=self.xge.dims_rec + self.xge.dims_n,
                name="labels",
            )
            .assign_coords(**{self._concat_dim: index, **self.state_kws})
            .assign_attrs(**self.xge._standard_attrs)
        )  # yapf: disable

        if reset_index:
            out = out.reset_index(self._concat_dim)

        return out

    @classmethod
    def from_labels(
        cls,
        ref,
        labels,
        lnzs,
        features=None,
        include_boundary=False,
        labels_kws=None,
        check_features=True,
        **kwargs
    ):
        r"""
        Create from reference :class:`~lnpy.lnPiMasked` and labels array


        Parameters
        ----------
        ref : lnPiMasked
        labels : sequence of label arrays
            Each `labels[i]` will be used to construct multiple phases from single
            (reweighted)  :math:`ln \Pi(N)`
        lnzs : sequence
            Each lnzs[i] will be passed to ``ref.reweight``.
        features : int, optional
        include_boundary : bool, default=False
        labels_kws : mapping, optional
        check_features : bool, default = True
        **kwargs
            Extra arguments past to :meth:`from_list`

        See Also
        --------
        labels_to_masks
        from_list
        """

        if labels_kws is None:
            labels_kws = {}
        assert len(labels) == len(lnzs)

        items = []
        indexes = []

        for label, lnz in zip(labels, lnzs):
            lnpi = ref.reweight(lnz)

            masks, features_tmp = labels_to_masks(
                labels=label,
                features=features,
                include_boundary=include_boundary,
                convention=False,
                check_features=check_features,
                **labels_kws
            )

            index = list(np.array(features_tmp) - 1)
            items += lnpi.list_from_masks(masks, convention=False)
            indexes += index

        return cls.from_list(items=items, index=indexes, **kwargs)

    @classmethod
    def from_dataarray(
        cls,
        ref,
        da,
        grouper="sample",
        include_boundary=False,
        labels_kws=None,
        features=None,
        check_features=True,
        **kwargs
    ):
        """
        Create a collection from DataArray of labels

        Parameters
        ----------
        ref : lnPiMasked
        da : DataArray or labels
        grouper : Hashable
            Name of dimension(s) to group along to give a single label array


        See Also
        --------
        from_labels


        """

        labels = []
        lnzs = []

        for i, g in da.groupby(grouper):
            lnzs.append(np.array([g.coords[k] for k in da.attrs["dims_lnz"]]))
            labels.append(g.values)

        return cls.from_labels(
            ref=ref,
            labels=labels,
            lnzs=lnzs,
            labels_kws=labels_kws,
            features=features,
            include_boundary=include_boundary,
            check_features=check_features,
            **kwargs
        )  # yapf: disable


################################################################################
# Accessors for ColleectionlnPi
@SeriesWrapper.decorate_accessor("zloc")
class _LocIndexer_unstack_zloc(object):
    """positional indexer for everything but phase"""

    def __init__(self, parent, level=["phase"]):
        self._parent = parent
        self._level = level
        self._loc = self._parent._series.unstack(self._level).iloc

    def __getitem__(self, idx):
        out = self._loc[idx]
        if isinstance(out, pd.DataFrame):
            out = out.stack(self._level)
        else:
            out = out.dropna()

        if isinstance(out, pd.Series):
            out = self._parent.new_like(out)
        return out


@SeriesWrapper.decorate_accessor("mloc")
class _LocIndexer_unstack_mloc(object):
    """indexer with pandas index"""

    def __init__(self, parent, level=["phase"]):
        self._parent = parent
        self._level = level
        self._index = self._parent.index

        self._index_names = set(self._index.names)
        self._loc = self._parent._series.iloc

    def _get_loc_idx(self, idx):
        index = self._index
        if isinstance(idx, pd.MultiIndex):
            # names in idx and
            drop = list(self._index_names - set(idx.names))
            index = index.droplevel(drop)
            # reorder idx
            idx = idx.reorder_levels(index.names)
        else:
            drop = list(set(index.names) - {idx.name})
            index = index.droplevel(drop)
        indexer = index.get_indexer_for(idx)
        return indexer

    def __getitem__(self, idx):
        indexer = self._get_loc_idx(idx)
        out = self._loc[indexer]

        if isinstance(out, pd.Series):
            out = self._parent.new_like(out)
        return out


# play around with sample index
# lnz_0 = np.random.rand(20)

# L = []
# for z in lnz_0:
#     repeat = np.random.randint(1, 4)
#     L += [(z, i) for i in range(repeat)]

# # example dataset
# idx = pd.MultiIndex.from_tuples(L, names=['lnz_0','phase'])
# s = pd.Series(range(len(L)), index=idx)

# def get_sample_index(s):
#     # get mapping from row values to sample
#     idx = s.index
#     idx_less_phase = idx.droplevel('phase')
#     idx_less_phase_unique = idx_less_phase.drop_duplicates()

#     mapping = pd.Series(range(len(idx_less_phase_unique)), idx_less_phase_unique)

#     # # sample values
#     # samples = mapping.loc[idx_less_phase].values

#     # make new index
#     idx_samp = (
#     idx
#     .to_frame()
#     .assign(sample=mapping.loc[idx_less_phase].values)
#     .set_index(['sample'] + idx.names)
#     .index
#     )
#     return idx_samp



"""
routines to find constant molfracs
"""

import numpy as np
from scipy import optimize

# from .segment import get_default_PhaseCreator


def _initial_bracket_molfrac(
    target,
    C,
    build_phases,
    phase_id=0,
    component=None,
    dlnz=0.5,
    dfac=1.0,
    ref=None,
    build_kws=None,
    ntry=20,
):
    """
    find bracket for molfrac
    """

    lnz_idx = build_phases.index
    if component is None:
        component = lnz_idx
    if build_kws is None:
        build_kws = {}

    if isinstance(phase_id, str) and phase_id.lower() == "none":
        # select stable phase
        skip_phase_id = True

        def getter(x):
            v = x.xge.molfrac.sel(component=component)
            if len(x) == 1:
                return v.isel(phase=0)
            else:
                return v.where(x.xge.mask_stable).max("phase")

    else:
        skip_phase_id = False

        def getter(x):
            return x.xge.molfrac.sel(component=component, phase=phase_id)

    s = getter(C).to_series().dropna()

    # get left bound
    left = None
    ntry_left = 0
    ss = s[s < target]
    if len(ss) > 0:
        left = C.mloc[ss.index[[-1]]]
    else:
        if len(s) > 0:
            index = s.index[[0]]
        else:
            index = C.index[[0]].droplevel("phase")
        new_lnz = C.mloc[index]._get_lnz(lnz_idx)
        dlnz_ = dlnz
        for i in range(ntry):
            new_lnz -= dlnz_
            dlnz_ *= dfac
            p = build_phases(new_lnz, ref=ref, **build_kws)
            if (skip_phase_id or phase_id in p._get_level("phase")) and getter(
                p
            ).values < target:
                left = p
                break
        ntry_left = i

    if left is None:
        raise RuntimeError("could not find left bounds")

    # right bracket
    right = None
    ntry_right = 0
    ss = s[s > target]
    if len(ss) > 0:
        right = C.mloc[ss.index[[0]]]
    else:
        if len(s) > 0:
            index = s.index[[-1]]
        else:
            index = C.index[[-1]].droplevel("phase")
        new_lnz = C.mloc[index]._get_lnz(lnz_idx)
        dlnz_ = dlnz

        for i in range(ntry):
            new_lnz += dlnz_
            p = build_phases(new_lnz, ref=ref, **build_kws)
            if (not skip_phase_id) and (phase_id not in p._get_level("phase")):
                # went to far
                new_lnz -= dlnz_
                # reset to half dlnz
                dlnz_ = dlnz_ * 0.5
            elif getter(p).values > target:
                right = p
                break
            else:
                dlnz_ *= dfac
        ntry_right = i

    if right is None:
        raise RuntimeError("could not find right bounds")

    info = dict(ntry_left=ntry_left, ntry_right=ntry_right)
    return left, right, info


def _solve_lnz_molfrac(
    target,
    left,
    right,
    build_phases,
    phase_id=0,
    component=None,
    build_kws=None,
    ref=None,
    tol=1e-4,
    **kwargs
):
    """
    calculate lnz which provides lnpi.molfracs_phaseIDs[phaseID,comp]==target

    Parameters
    ----------
    target : float
        target molfraction
    a, b : float
        lnz values bracketing solution
    phase_id : int
        target phase
    ref : lnPiMasked
        object to reweight
    component : int, optional
        if not specified, use build_phases.index
    full_output : bool (Default False)
        if True, return solve stats
    tol : float (default 1e-4)
        solver tolerance

    **kwargs : extra arguments to scipy.optimize.brentq

    Returns
    --------
    output : lnPi_phases object
        object with desired molfraction
    info : solver info (optional, returned if full_output is `True`)
    """

    if build_kws is None:
        build_kws = {}

    if component is None:
        component = build_phases.index

    if not isinstance(left, float):
        left = left._get_lnz(build_phases.index)
    if not isinstance(right, float):
        right = right._get_lnz(build_phases.index)

    a, b = sorted([x for x in (left, right)])

    if isinstance(phase_id, str) and phase_id.lower() == "none":
        # select stable phase
        skip_phase_id = True

        def getter(x):
            v = x.xge.molfrac.sel(component=component)
            if len(x) == 1:
                return v.isel(phase=0)
            else:
                return v.where(x.xge.mask_stable).max("phase")

    else:
        skip_phase_id = False

        def getter(x):
            return x.xge.molfrac.sel(component=component, phase=phase_id)

    def f(x):
        p = build_phases(x, ref=ref, **build_kws)
        f.lnpi = p

        # by not using the ListAccessor,
        # can parralelize
        if skip_phase_id or phase_id in p._get_level("phase"):
            mf = getter(p).values
            # mf = (
            #     p.s.xs(phase_id, level='phase').iloc[0]
            #     .xge
            #     .molfrac
            #     .sel(component=component)
            #     .values
            # )
        else:
            mf = np.inf

        return mf - target

    xx, r = optimize.brentq(f, a, b, full_output=True, **kwargs)
    r.residual = f(xx)

    if np.abs(r.residual) > tol:
        raise RuntimeError("something went wrong with solve")
    return f.lnpi, r


def find_lnz_molfrac(
    target,
    C,
    build_phases,
    phase_id=0,
    component=None,
    build_kws=None,
    ref=None,
    dlnz=0.5,
    dfac=1.0,
    ntry=20,
    tol=1e-4,
    **kwargs
):

    left, right, info = _initial_bracket_molfrac(
        target=target,
        C=C,
        build_phases=build_phases,
        phase_id=phase_id,
        component=component,
        dlnz=dlnz,
        dfac=dfac,
        ref=ref,
        build_kws=build_kws,
        ntry=ntry,
    )

    lnpi, r = _solve_lnz_molfrac(
        target=target,
        left=left,
        right=right,
        build_phases=build_phases,
        phase_id=phase_id,
        component=component,
        build_kws=build_kws,
        ref=ref,
        tol=tol,
        **kwargs
    )
    return lnpi, r



from .ensembles import xCanonical, xGrandCanonical
from .lnpidata import MaskedlnPiDelayed, lnPiArray, lnPiMasked
from .lnpienergy import (
    merge_regions,
    wFreeEnergy,
    wFreeEnergyCollection,
    wFreeEnergyPhases,
)
from .lnpiseries import lnPiCollection
from .maskedlnpi_legacy import MaskedlnPiLegacy
from .options import OPTIONS, set_options
from .segment import (
    BuildPhases_dmu,
    BuildPhases_mu,
    PhaseCreator,
    Segmenter,
    peak_local_max_adaptive,
)
from .stability import Binodals, Spinodals
from .utils import dim_to_suffix

# Legacy stuff


# updated versioning scheme
try:
    from importlib.metadata import version as _version
except ImportError:
    # if the fallback library is missing, we are doomed.
    from importlib_metadata import version as _version  # type: ignore[no-redef]


try:
    __version__ = _version("lnpy")
except Exception:
    # Local copy or not installed with setuptools.
    # Disable minimum version checks on downstream libraries.
    __version__ = "999"


__author__ = """William P. Krekelberg"""
__email__ = "wpk@nist.gov"


__all__ = [
    "__author__",
    "__email__",
    "__version__",
    "lnPiArray",
    "lnPiMasked",
    "lnPiCollection",
    "MaskedlnPiDelayed",
    "MaskedlnPiLegacy",
    "xGrandCanonical",
    "xCanonical",
    "wFreeEnergyPhases",
    "wFreeEnergyCollection",
    "wFreeEnergy",
    "merge_regions",
    "Segmenter",
    "PhaseCreator",
    "peak_local_max_adaptive",
    "BuildPhases_mu",
    "BuildPhases_dmu",
    "dim_to_suffix",
    "set_options",
    "OPTIONS",
    "Spinodals",
    "Binodals",
]



"""
thermodynamic stability routines (i.e., spinodal/binodal)
"""
import itertools

import numpy as np
from scipy import optimize

from .cached_decorators import gcached
from .lnpiseries import lnPiCollection


# ###############################################################################
# Spinodal routines
def _initial_bracket_spinodal_right(
    C,
    build_phases,
    idx,
    idx_nebr=None,
    efac=1.0,
    dlnz=0.5,
    dfac=1.0,
    vmax=1e5,
    ntry=20,
    step=+1,
    ref=None,
    build_kws=None,
):
    """
    find initial bracketing lnpi_phases of phaseID==ID bracketing point where
    DeltabetaE_phaseIDS()[ID]==efac

    Parameters
    ----------
    ref : lnPiMasked, optional
    build_phases : callable
        scalar funciton to build phases
    C : lnPi_collection
        initial estimates to work from
    idx, idx_nebr : int
        id's of from/to phases.
    lnz_in : list
        list with value of static chem pot, and None for variable. e.g.,
        lnz_in=[None,0.0] implies lnz[0] is variable, and lnz[1]=0.0
    efac : float (Default 1.0)
        cutoff value for spinodal
    dlnz : float (Default 0.5)
        factor to kick back if C doesn't already have left and right bounds
    vmax : float (default 1e20)
        value indicating no transition, but phaseID==ID present
    ntry : int (Default 20)
        number of times to try kicking forward/backwards to find bracket
    step : int (Default +1)
        if +1, step forward
        if -1, step backward
    build_phases : callable, optional.
        function that returns a Phases object.
    build_kws : dict
        extra arguments to build_phases
    Returns
    -------
    left,right: lnpi_phases
        left and right bracketing lnPi_phases objects

    """

    if build_kws is None:
        build_kws = {}
    if efac <= 0:
        raise ValueError("efac must be positive")

    # Series representation of dw
    s = C.wfe.get_dw(idx, idx_nebr)
    if step < 0:
        s = s.iloc[-1::-1]

    # see if contain "left" bounds
    left = None
    s_idx = s[s > 0.0]
    if len(s_idx) == 0:
        raise ValueError("no phase {}".format(idx))

    ss = s_idx[s_idx > efac]
    if len(ss) > 0:
        # get last one
        left = C.mloc[ss.index[[-1]]]
    else:
        new_lnz = C.mloc[s_idx.index[[0]]]._get_lnz(build_phases.index)
        for i in range(ntry):
            new_lnz -= step * dlnz
            t = build_phases(new_lnz, ref=ref, **build_kws)
            if (
                idx in t._get_level("phase")
                and t.wfe_phases.get_dw(idx, idx_nebr) > efac
            ):
                left = t
                break

    if left is None:
        raise RuntimeError("could not find left")

    # right
    right = None
    ss = s[s < efac]
    if len(ss) > 0:
        right = C.mloc[ss.index[[0]]]
    else:
        new_lnz = C.mloc[s.index[[-1]]]._get_lnz(build_phases.index)
        dlnz_ = dlnz
        for i in range(ntry):
            new_lnz += step * dlnz_
            t = build_phases(new_lnz, ref=ref, **build_kws)
            if (
                idx not in t._get_level("phase")
                or t.wfe_phases.get_dw(idx, idx_nebr) < efac
            ):
                right = t
                break
            dlnz_ *= dfac

    if right is None:
        raise RuntimeError("could not find right")
    return left, right


def _refine_bracket_spinodal_right(
    left,
    right,
    build_phases,
    idx,
    idx_nebr=None,
    efac=1.0,
    nmax=30,
    vmax=1e5,
    vmin=0.0,
    ref=None,
    build_kws=None,
    close_kws=None,
):
    """
    find refined bracket with efac<DeltabetaE_left<vmax and
    vmin<DeltabetaE_right<efac

    Parameters
    ----------
    left, right : Phases objects
        left and right initial brackets
    idx, idx_nebr : int
        from/to phase id's
    efac : float (Default 1.0)
        cutoff value for spinodal
    nmax : int (Default 30)
        max number of interations
    vmin,vmax : see above
    build_phases : callable
    build_kws : dict
    close_kwargs : dict
        arguments to np.allclose
    Returns
    -------
    left,right : lnpi_phases objects
        left and right phases bracketing spinodal

    r : scipy.optimize.zeros.RootResults object
    """

    doneLeft = False
    doneRight = False
    if build_kws is None:
        build_kws = {}
    if close_kws is None:
        close_kws = {}

    for i in range(nmax):
        # if idx in left.index and idx_nebr in left.index:
        # dw = _get_dw(left, idx, idx_nebr)
        dw = left.wfe_phases.get_dw(idx, idx_nebr)
        if dw < vmax and dw > efac:
            doneLeft = True

        # dw = _get_dw(right, idx, idx_nebr)
        dw = right.wfe_phases.get_dw(idx, idx_nebr)
        if dw > vmin and dw < efac:
            doneRight = True

        #########
        # checks
        if doneLeft and doneRight:
            # find bracket
            r = optimize.RootResults(root=None, iterations=i, function_calls=i, flag=1)
            return left, right, r

        ########
        # converged?
        if np.allclose(left._get_lnz(), right._get_lnz(), **close_kws):
            # we've reached a breaking point
            if doneLeft:
                # can't find a lower bound to efac, just return where we're at
                r = optimize.RootResults(
                    root=left._get_lnz(), iterations=i + 1, function_calls=i, flag=0
                )
                for k, val in [
                    ("left", left),
                    ("right", right),
                    ("doneleft", doneLeft),
                    ("doneright", "doneRight"),
                    ("info", "all close and doneleft"),
                ]:
                    setattr(r, k, val)
                return left, right, r

            # elif not doneLeft and not doneRight:
            else:
                # all close, and no good on either end -> no spinodal
                r = optimize.RootResults(
                    root=None, iterations=i + 1, function_calls=i, flag=1
                )
                for k, val in [
                    ("left", left),
                    ("right", right),
                    ("doneleft", doneLeft),
                    ("doneright", "doneRight"),
                    ("info", "all close and doneleft"),
                ]:
                    setattr(r, k, val)
                for k, val in [
                    ("left", left),
                    ("right", right),
                    ("info", "all close and not doneleft"),
                ]:
                    setattr(r, k, val)
                return None, None, r

        # mid point phases
        lnz_mid = 0.5 * (
            left._get_lnz(build_phases.index) + right._get_lnz(build_phases.index)
        )

        mid = build_phases(lnz_mid, ref=ref, **build_kws)
        dw = mid.wfe_phases.get_dw(idx, idx_nebr)

        if (idx in mid._get_level("phase")) and (
            mid.wfe_phases.get_dw(idx, idx_nebr) >= efac
        ):
            left = mid
        else:
            right = mid

    raise RuntimeError(
        f"""
    did not finish
    ntry      : {i}
    idx       : {idx}
    idx_nebr  : {idx_nebr}
    left lnz   : {left._get_lnz()}
    right lnz  : {right._get_lnz()}
    doneleft  : {doneLeft}
    doneright : {doneRight}
    """
    )


def _solve_spinodal(
    a, b, build_phases, idx, idx_nebr=None, efac=1.0, ref=None, build_kws=None, **kwargs
):
    if build_kws is None:
        build_kws = {}

    def f(x):
        c = build_phases(x, ref=ref, **build_kws)
        dw = c.wfe_phases.get_dw(idx, idx_nebr)
        out = dw - efac

        f._lnpi = c
        f._out = out

        return out

    xx, r = optimize.brentq(f, a, b, full_output=True, **kwargs)

    r.residual = f(xx)
    lnz = f._lnpi._get_lnz(build_phases.index)
    return lnz, r, f._lnpi


def _get_step(C, idx, idx_nebr):
    """
    find step value on

    if DeltabetaE[-1,ID] - DeltabetaE[0,ID]<0 -> step=+1 (go right)

    else step = -1
    """
    delta = C.zloc[[-1]].wfe_phases.get_dw(idx, idx_nebr) - C.zloc[
        [0]
    ].wfe_phases.get_dw(idx, idx_nebr)
    if delta == 0:
        raise ValueError("could not determine step, delta==0")
    elif delta < 0.0:
        return +1
    else:
        return -1


def get_spinodal(
    C,
    build_phases,
    idx,
    idx_nebr=None,
    efac=1.0,
    dlnz=0.5,
    dfac=1.0,
    vmin=0.0,
    vmax=1e5,
    ntry=20,
    step=None,
    nmax=20,
    ref=None,
    build_kws=None,
    close_kws=None,
    solve_kws=None,
    full_output=False,
):
    """
        locate spinodal point for a given phaseID
    _
        Parameters
        ----------
        ref : lnPiMasked
        C : lnPi_collection
            initial estimates to work from.  Function assumes C is in lnz sorted order
        idx, idx_nebr : int
            from/to phase id
        lnz_in : list
            list with value of static chem pot, and None for variable. e.g.,
            lnz_in=[None,0.0] implies lnz[0] is variable, and lnz[1]=0.0
        efac : float, optional
            cutoff value for spinodal
        dlnz : float, optional
            factor to kick back if C doesn't already have left and right bounds
        vmin : float, optional
            value denoting vmin, i.e., value of DeltabetaE if phaseID does not exist
        vmax : float, optional
            value indicating no transition, but phaseID==ID present
        ntry : int (Default 20)
            number of times to try kicking forward/backwards to find bracket
        step : int or None (Default None)
            if +1, step forward
            if -1, step backward
            if None, try to determine step
        nmax : int (Default 20)
            max number of steps to refine bracket
        build_phases : callable, optional
            function to create Phases.  Default is that fro get_default_PhaseCreator
        build_kws : dict, optional
            extra arguments to `build_phases`
        close_kws : dict, optional
            arguments to np.allclose
        solve_kws : dict, optional
            extra arguments to scipy.optimize.brentq
        full_output : bool (Default False)
            if true, return output info object

        Returns
        -------
        out : lnPi_phases object at spinodal point
        r :output info object (optional, returned if full_output is True)

    """
    assert len(C) > 1
    if build_kws is None:
        build_kws = {}
    if close_kws is None:
        close_kws = {}
    if solve_kws is None:
        solve_kws = {}

    if step is None:
        step = _get_step(C, idx=idx, idx_nebr=idx_nebr)

    assert step in [-1, +1]

    # if step == +1:
    #     CC = C
    # elif step == -1:
    #     CC = C[-1::-1]
    # else:
    #     raise ValueError('bad step')

    # get initial bracket
    L, R = _initial_bracket_spinodal_right(
        C,
        idx=idx,
        idx_nebr=idx_nebr,
        efac=efac,
        dlnz=dlnz,
        dfac=dfac,
        vmax=vmax,
        ntry=ntry,
        step=step,
        ref=ref,
        build_phases=build_phases,
        build_kws=build_kws,
    )

    left, right, rr = _refine_bracket_spinodal_right(
        L,
        R,
        idx=idx,
        idx_nebr=idx_nebr,
        efac=efac,
        nmax=nmax,
        vmin=vmin,
        vmax=vmax,
        ref=ref,
        build_phases=build_phases,
        build_kws=build_kws,
        close_kws=close_kws,
    )

    if left is None and right is None:
        # no spinodal found and left and right are close
        spin = None
        r = rr
    elif rr.converged:
        # converged to a solution
        spin = left
        r = rr
        r.bracket_iteration = rr.iterations
        r.from_solve = False
    else:
        # solve
        if step == -1:
            left, right = right, left

        a, b = [x._get_lnz(build_phases.index) for x in [left, right]]
        lnz, r, spin = _solve_spinodal(
            ref=ref,
            idx=idx,
            idx_nebr=idx_nebr,
            a=a,
            b=b,
            efac=efac,
            build_phases=build_phases,
            build_kws=build_kws,
            **solve_kws,
        )
        r.bracket_iterations = rr.iterations
        r.from_solve = True
    if full_output:
        return spin, r
    else:
        return spin


################################################################################
# Binodal routines


def get_binodal_point(
    IDs,
    lnz_min,
    lnz_max,
    build_phases,
    ref=None,
    build_kws=None,
    full_output=False,
    **kwargs,
):
    """
    calculate binodal point where Omega[ID[0]]==Omega[ID[1]]

    Parameters
    ----------
    ref : lnPi
        object to reweight
    IDs : tuple
        phase index of pair to equate
    lnz_min, lnz_max : float
        lnz_index values bracketing solution
    build_phases : callable
        function to create Phases object
    build_kws : dict, optional
        optional arguments to build_phases
    full_output : bool (Default False)
        if True, return solve stats
    kwargs : dict
        extra arguments to scipy.optimize.brentq
    Returns
    -------
    binodal : Phases object at binodal point
    stats : optional,
        solve stats object from brentq (optional, returned if full_output is True)
    """

    IDs = list(IDs)
    assert len(IDs) == 2

    if build_kws is None:
        build_kws = {}

    # lnz_idx = build_phases.index

    a, b = min(lnz_min, lnz_max), max(lnz_min, lnz_max)

    def f(x):
        p = build_phases(x, ref=ref, **build_kws)
        f.lnpi = p
        # Omegas = c.omega_phase()
        # return Omegas[IDs[0]] - Omegas[IDs[1]]
        return p.xge.betaOmega().reindex(phase=IDs).diff("phase").squeeze().values

    xx, r = optimize.brentq(f, a, b, full_output=True, **kwargs)
    r.residual = f(xx)
    if full_output:
        return f.lnpi, r
    else:
        return f.lnpi


################################################################################
# Accessor classes/routines


class _BaseStability(object):
    """

    Parameters
    ----------
    collection: lnPiCollection
        Used to bracket the location limit of stability


    Methods
    -------
    __call__ : set value
    """

    _NAME = "base"

    def __init__(self, collection):
        self._parent = collection
        self.access_kws = {}

    @property
    def parent(self):
        return self._parent

    def set_values(self, items, info, index=None):
        self._items = items
        self._info = info
        self._index = index

    def set_access_kws(self, **kwargs):
        for k, v in kwargs.items():
            self.access_kws[k] = v

    @property
    def items(self):
        """Access to the underlying data"""
        return self._items

    def _get_access(self, items=None, concat_kws=None, **kwargs):
        if items is None:
            items = self._items
        if concat_kws is None:
            concat_kws = {}
        concat_kws = dict(names=[self._NAME], **concat_kws)
        kwargs = dict(self.access_kws, **kwargs)
        # return self._parent.concat_like(items, **concat_kws)
        # s = pd.concat({k:v._series for k,v in items}, **concat_kws)
        # return self._parent.new_like(s)
        return self._parent.concat(items, concat_kws=concat_kws, **kwargs)

    @gcached()
    def access(self):
        """lnPiCollection view of stability"""
        return self._get_access()

    def __getitem__(self, idx):
        return self._items[idx]

    def _get_appender(self, s=None):
        if s is None:
            s = self.access
        return s.droplevel(self._NAME)

    @property
    def appender(self):
        return self._get_appender()

    def append_to_parent(self, sort=True, copy_stability=True):
        new = self._parent.append(self.appender)
        if sort:
            new = new.sort_index()
        if copy_stability:
            setattr(new, self._NAME, self)
        return new


# NOTE : single create means this is only created once
@lnPiCollection.decorate_accessor("spinodal", single_create=False)
class Spinodals(_BaseStability):
    """
    Methods for calculation locations of spinodal
    """

    _NAME = "spinodal"

    def __call__(
        self,
        phase_ids,
        build_phases,
        efac=1.0,
        ref=None,
        build_kws=None,
        inplace=True,
        # append=False,
        force=False,
        as_dict=True,
        unstack=None,
        **kwargs,
    ):
        """
        Parameters
        ----------
        phase_ids : int, or sequence
            Phase ids to calculate spinodal for.  If int, then find spinodals for phases
            range(phase_ids).
        build_phases : callable
            Factory function to build phases.
            This should most likely be an instance of :class:`lnpy.BuildPhases_mu`
        efac : float, default=1.0
            Target value of `dw` to define spinodal.
        ref : lnPiMasked, optional
        build_kws : dict, optional
            optional arguments to `build_phases`
        inplace : bool, default=True
            if True, add spinodal inplace, otherwise return spinodal object
        force : bool, optional
            if True, force recalcuation of spinodal if alread set inplace.  Otherwise return already
            calculated values
        as_dict : bool, default=True
            if True, return dict of form {phase_id[0] : phases object, ...}.
        unstack : bool, optional
            if passed, create lnPiCollection objects with this unstack value.  If not passed,
            use unstack parameter from parent object
        **kwargs : extra argument to `get_spinodal` function

        Returns
        -------
        out : output
            if inplace, return self.
            if not inplace, and as dict, return dict, else return :class:`lnpy.lnPiCollection` with phase_id in index
        """

        if hasattr(self, "_items") and not force:
            if inplace:
                return self
            else:
                if as_dict:
                    out = self._items
                else:
                    out = self.access
                return out, self._info

        from .segment import BuildPhasesBase

        if not isinstance(build_phases, BuildPhasesBase):
            raise ValueError(
                "`build_phases` should be an instance of `BuildPhasesBase`."
                "Its likely an instance of `PhaseCreator.builphases`."
                "Instead, use an instance of `PhaseCreator.buildphases_mu`."
            )

        if isinstance(phase_ids, int):
            phase_ids = list(range(phase_ids))
        if not isinstance(phase_ids, list):
            raise ValueError("phase_ids must be an int or list")

        out = {}
        info = {}
        kwargs["full_output"] = True
        for idx in phase_ids:
            s, r = get_spinodal(
                ref=ref,
                C=self._parent,
                idx=idx,
                build_phases=build_phases,
                build_kws=build_kws,
                **kwargs,
            )
            out[idx] = s
            info[idx] = r

        if unstack is None:
            unstack = self._parent._xarray_unstack
        self.set_access_kws(unstack=unstack)

        if inplace:
            self._items = out
            self._info = info
            # if append:
            #     return self._append_to_parent()
            # else:
            return self
        else:
            if not as_dict:
                out = self._get_access(out)
            return out, info


@lnPiCollection.decorate_accessor("binodal", single_create=False)
class Binodals(_BaseStability):
    """Routines to calculate binodal."""

    _NAME = "binodal"

    def get_pair(
        self,
        ids,
        lnz_min=None,
        lnz_max=None,
        spinodals=None,
        ref=None,
        build_phases=None,
        build_kws=None,
        **kwargs,
    ):

        if None in [lnz_min, lnz_max] and spinodals is None:
            spinodals = self._parent.spinodal
        if lnz_min is None:
            lnz_min = spinodals[ids[0]]._get_lnz(build_phases.index)
        if lnz_max is None:
            lnz_max = spinodals[ids[1]]._get_lnz(build_phases.index)
        return get_binodal_point(
            ref=ref,
            IDs=ids,
            lnz_min=lnz_min,
            lnz_max=lnz_max,
            build_phases=build_phases,
            build_kws=build_kws,
            **kwargs,
        )

    def __call__(
        self,
        phase_ids,
        build_phases,
        spinodals=None,
        ref=None,
        build_kws=None,
        inplace=True,
        force=False,
        as_dict=True,
        unstack=None,
        **kwargs,
    ):
        """
        Parameters
        ----------
        phase_ids : int, or sequence
            Phase ids to calculate binodal for.  If int, then find binodal for phases
            range(phase_ids).
        build_phases : callable
            Factory function to build phases
        spinodals : optional
            if not passes, then use parent.spinodal
            Used for bounding binodal
        ref : lnPiMasked, optional
        build_kws : dict, optional
            optional arguments to `build_phases`
        inplace : bool, default=True
            if True, add spinodal inplace, otherwise return spinodal object
        force : bool, optional
            if True, force recalcuation of spinodal if alread set inplace.  Otherwise return already
            calculated values
        as_dict : bool, default=True
            if True, return dict of form {phase_id[0] : phases object, ...}
        unstack : bool, optional
            if passed, create lnPiCollection objects with this unstack value.  If not passed,
            use unstack parameter from parent object
        **kwargs : extra argument to `get_spinodal` function

        Returns
        -------
        out : output
            if inplace, return self
            if not inplace, and as dict, return dict, else return lnPiCollection with phase_id in index
        """

        if inplace and hasattr(self, "_items") and not force:
            if inplace:
                # if append:
                #     if append_kws is None:
                #         append_kws = {}
                #     return self.append_to_parent(**append_kws)
                # else:
                return self
            else:
                if as_dict:
                    out = self._items
                else:
                    out = self.access
                return out, self._info

        if isinstance(phase_ids, int):
            phase_ids = list(range(phase_ids))
        if not isinstance(phase_ids, list):
            raise ValueError("phase_ids must be an int or list")

        out = {}
        info = {}
        index = {}
        kwargs["full_output"] = True
        for idx, ids in enumerate(itertools.combinations(phase_ids, 2)):
            s, r = self.get_pair(
                ref=ref,
                ids=ids,
                spinodals=spinodals,
                build_phases=build_phases,
                build_kws=build_kws,
                **kwargs,
            )
            out[idx] = s
            info[idx] = r
            index[idx] = ids

        if unstack is None:
            unstack = self._parent._xarray_unstack
        self.set_access_kws(unstack=unstack)

        # either build output or inplace
        if inplace:
            self._items = out
            self._info = info
            self._index = index
            # if append:
            #     return self.append_to_parent()
            # else:
            return self
        else:
            if not as_dict:
                out = self._get_access(out)
            return out, info


@lnPiCollection.decorate_accessor("stability_append", single_create=False)
def _stability_append(self):
    """
    add stability from collection to this collection

    Parameters
    ----------
    other : optional, default=self
        if passed, copy stability from this collection to self, otherwise
        use self
    append: bool, default=True
        if True, append results to new frame
    sort: bool, default=True
        if True, sort appended results
    copy_stability
    """
    # Duplicate docstring so show up in docs

    def func(other=None, append=True, sort=True, copy_stability=True):
        """
        add stability from collection to this collection

        Parameters
        ----------
        other : optional, default=self
            if passed, copy stability from this collection to self, otherwise
            use self
        append: bool, default=True
            if True, append results to new frame
        sort: bool, default=True
            if True, sort appended results
        copy_stability
        """

        if (not append) and (not copy_stability):
            raise ValueError("one of append or copy_stability must be True")

        if other is None:
            other = self
        spin = other.spinodal
        bino = other.binodal
        if append:
            new = self.append(spin.appender).append(bino.appender)
            if sort:
                new = new.sort_index()
        else:
            new = self.copy()
        if copy_stability:
            new.spinodal = spin
            new.binodal = bino
        return new

    return func



from __future__ import absolute_import, division, print_function

from warnings import warn

import numpy as np
import pandas as pd
from scipy.ndimage import filters

from .cached_decorators import gcached
from .extensions import AccessorMixin
from .utils import labels_to_masks, masks_change_convention

# NOTE : This is a rework of core.
# [ ] : split xarray functionality into wrapper(s)
# [ ] : split splitting into separate classes


class MaskedlnPiLegacy(np.ma.MaskedArray, AccessorMixin):
    """
    class to store masked ln[Pi(n0,n1,...)].
    shape is (N0,N1,...) where Ni is the span of each dimension)

    Attributes
    ----------
    self : masked array containing lnPi
    lnz : log(absolute activity) = beta * (chemical potential) for each component
    coords : coordinate array (ndim,N0,N1,...)
    pi : exp(lnPi)
    grand : grand potential (-pV) of system
    argmax_local : local argmax (in np.where output form)
    zeromax : set lnPi = lnPi - lnPi.max()
    pad : fill in masked points by interpolation
    adjust : zeromax and/or pad
    reweight : create new lnPi at new mu
    smooth : create smoothed object
    """

    def __new__(cls, data=None, lnz=None, state_kws=None, extra_kws=None, **kwargs):
        """
        constructor

        Parameters
        ----------
        data : array-like
         data for lnPi

        lnz : array-like (Default None)
            if None, set lnz=np.zeros(data.ndim)
        state_kws : dict, optional
            dictionary of state values, such as `volume` and `beta`.
            These parameters will be pushed to `self.xge` coordinates.
        extra_kws : dict, optional
            this defines extra parameters to pass along.
            Note that for potential energy calculations, extra_kws should contain
            `PE` (total potentail energy for each N vector).
        zeromax : bool (Default False)
            if True, shift lnPi = lnPi - lnPi.max()
        pad : bool (Default False)
            if True, pad masked region by interpolation
        kwargs : arguments to np.ma.array
            e.g., mask=...
        """
        warn("MaskedlnPiLegacy is deprecated.  Please use lnPiMasked instead")

        if data is not None and issubclass(data.dtype.type, np.floating):
            kwargs.setdefault("fill_value", np.nan)

        obj = np.ma.array(data, **kwargs).view(cls)
        # fv = kwargs.get('fill_value', None) or getattr(data, 'fill_value', None)
        # if fv is None:
        #     fv = np.nan
        # obj.set_fill_value(fv)

        # make sure to broadcase mask if it is just False
        if obj.mask is False:
            obj.mask = False

        # set mu value:
        if lnz is None:
            lnz = np.zeros(obj.ndim)
        lnz = np.atleast_1d(lnz).astype(obj.dtype)
        if len(lnz) != obj.ndim:
            raise ValueError("bad len on lnz %s" % lnz)

        if state_kws is None:
            state_kws = {}
        if extra_kws is None:
            extra_kws = {}

        obj._optinfo.update(
            lnz=lnz,
            state_kws=state_kws,
            extra_kws=extra_kws,
        )
        return obj

    ##################################################
    # caching
    def __array_finalize__(self, obj):
        super().__array_finalize__(obj)
        self._clear_cache()

    def _clear_cache(self):
        self._cache = {}

    ##################################################
    # properties
    @property
    def optinfo(self):
        """all extra properties"""
        return self._optinfo

    @property
    def state_kws(self):
        """state specific parameters"""
        return self._optinfo["state_kws"]

    @property
    def extra_kws(self):
        """all extra parameters"""
        return self._optinfo["extra_kws"]

    def _index_dict(self, phase=None):

        out = {"lnz_{}".format(i): v for i, v in enumerate(self.lnz)}
        if phase is not None:
            out["phase"] = phase
        # out.update(**self.state_kws)
        return out

    def _lnpi_tot(self, fill_value=None):
        return self.filled(fill_value)

    def _pi_params(self, fill_value=None):
        lnpi = self._lnpi_tot(fill_value)

        lnpi_local_max = lnpi.max()
        pi = np.exp(lnpi - lnpi_local_max)
        pi_sum = pi.sum()
        pi_norm = pi / pi_sum

        lnpi_zero = self.data.ravel()[0] - lnpi_local_max

        return pi_norm, pi_sum, lnpi_zero

    @property
    def _lnz_tot(self):
        return self.lnz

    # @property
    # def _lnpi_0_tot(self):
    #     return self.data.ravel()[0]

    @property
    def lnz(self):
        return self._optinfo.get("lnz", None)

    @property
    def betamu(self):
        return self.lnz

    # @property
    # def mu(self):
    #     return self._optinfo.get('mu', None)
    @property
    def volume(self):
        return self.state_kws.get("volume", None)

    @property
    def beta(self):
        return self.state_kws.get("beta", None)

    def __repr__(self):
        L = []
        L.append("lnz={}".format(repr(self.lnz)))
        L.append("state_kws={}".format(repr(self.state_kws)))

        L.append("data={}".format(super().__repr__()))
        if len(self.extra_kws) > 0:
            L.append("extra_kws={}".format(repr(self.extra_kws)))

        indent = " " * 5
        p = "MaskedlnPi(\n" + "\n".join([indent + x for x in L]) + "\n)"

        return p

    def __str__(self):
        return "MaskedlnPi(lnz={})".format(str(self.lnz))

    # @gcached(prop=False)
    def local_argmax(self, *args, **kwargs):
        return np.unravel_index(self.argmax(*args, **kwargs), self.shape)

    # @gcached(prop=False)
    def local_max(self, *args, **kwargs):
        return self[self.local_argmax(*args, **kwargs)]

    # @gcached(prop=False)
    def local_maxmask(self, *args, **kwargs):
        return self == self.local_max(*args, **kwargs)

    @gcached()
    def edge_distance_matrix(self):
        """matrix of distance from upper bound"""
        from .utils import distance_matrix

        return distance_matrix(~self.mask)

    def edge_distance(self, ref, *args, **kwargs):
        return ref.edge_distance_matrix[self.local_argmax(*args, **kwargs)]

    # make these top level
    # @gcached()
    # @property
    # def pi(self):
    #     """
    #     basic pi = exp(lnpi)
    #     """
    #     pi = np.exp(self - self.local_max())
    #     return pi

    # @gcached()
    # def pi_sum(self):
    #     return self.pi.sum()

    # @gcached(prop=False)
    # def betaOmega(self, lnpi_zero=None):
    #     if lnpi_zero is None:
    #         lnpi_zero = self.data.ravel()[0]
    #     zval = lnpi_zero - self.local_max()
    #     return  (zval - np.log(self.pi_sum))

    def __setitem__(self, index, value):
        self._clear_cache()
        super().__setitem__(index, value)

    def pad(self, axes=None, ffill=True, bfill=False, limit=None, inplace=False):
        """
        pad nan values in underlying data to values

        Parameters
        ----------
        ffill : bool, default=True
            do forward filling
        bfill : bool, default=False
            do back filling
        limit : int, default None
            The maximum number of consecutive NaN values to forward fill. In
            other words, if there is a gap with more than this number of
            consecutive NaNs, it will only be partially filled. Must be greater
            than 0 or None for no limit.
        inplace : bool, default=False

        Returns
        -------
        out : lnPi
            padded object
        """
        import bottleneck

        from .utils import bfill, ffill

        if axes is None:
            axes = range(self.ndim)

        data = self.data
        datas = []

        if ffill:
            datas += [ffill(data, axis=axis, limit=limit) for axis in axes]
        if bfill:
            datas += [bfill(data, axis=axis, limit=limit) for axis in axes]

        if len(datas) > 0:
            data = bottleneck.nanmean(datas, axis=0)

        if inplace:
            new = self
            new._clear_cache()
        else:
            new = self.copy()

        new.data[...] = data
        return new

    def zeromax(self, inplace=False):
        """
        shift so that lnpi.max() == 0
        """

        if inplace:
            new = self
            self._clear_cache()
        else:
            new = self.copy()

        new.data[...] = new.data - new.max()
        return new

    def adjust(self, zeromax=False, pad=False, inplace=False):
        """
        do multiple adjustments in one go
        """

        if inplace:
            new = self
        else:
            new = self.copy()

        if zeromax:
            new.zeromax(inplace=True)
        if pad:
            new.pad(inplace=True)
        return new

    def reweight(self, lnz, zeromax=False, pad=False):
        """
        get lnpi at new lnz

        Parameters
        ----------
        lnz : array-like
            chem. pot. for new state point

        zeromax : bool (Default False)

        pad : bool (Default False)

        phases : dict


        Returns
        -------
        lnPi(lnz)
        """

        lnz = np.atleast_1d(lnz)

        assert len(lnz) == len(self.lnz)

        new = self.copy()
        new._optinfo["lnz"] = lnz

        dlnz = new.lnz - self.lnz

        # s = _get_shift(self.shape,dmu)*self.beta
        # get shift
        # i.e., N * (mu_1 - mu_0)
        # note that this is (for some reason)
        # faster than doing the (more natural) options:
        # N = self.ncoords.values
        # shift = 0
        # for i, m in enumerate(dmu):
        #     shift += N[i,...] * m
        # or
        # shift = (self.ncoords.values.T * dmu).sum(-1).T

        shift = np.zeros([], dtype=float)
        for i, (s, m) in enumerate(zip(self.shape, dlnz)):
            shift = np.add.outer(shift, np.arange(s) * m)

        # scale by beta
        # shift *= self.beta

        new.data[...] += shift
        new.adjust(zeromax=zeromax, pad=pad, inplace=True)

        return new

    def smooth(self, sigma=4, mode="nearest", truncate=4, inplace=False, **kwargs):
        """
        apply gaussian filter smoothing to data

        Parameters
        ----------
        inplace : bool (Default False)
         if True, do inplace modification.
        mode, truncate : arguments to filters.gaussian_filter

        **kwargs : (Default sigma=4, mode='nearest',truncate=4)
         arguments to filters.gaussian_filter
        """

        if inplace:
            new = self
            new._clear_cache()
        else:
            new = self.copy()

        filters.gaussian_filter(
            new.data,
            output=new.data,
            mode=mode,
            truncate=truncate,
            sigma=sigma,
            **kwargs
        )
        return new

    def copy_shallow(self, mask=None, **kwargs):
        """
        create shallow copy

        Parameters
        ----------
        mask : optional
            if specified, new object has this mask
            otherwise, at least copy old mask
        """
        if mask is None:
            mask = self.mask.copy()

        return self.__class__(
            self.data,
            mask=mask,
            fill_value=self.fill_value,
            **dict(self._optinfo, **kwargs)
        )

    def or_mask(self, mask, **kwargs):
        """
        new object with logical or of self.mask and mask
        """
        return self.copy_shallow(mask=mask + self.mask, **kwargs)

    def and_mask(self, mask, **kwargs):
        """
        new object with logical and of self.mask and mask
        """
        return self.copy_shallow(mask=mask * self.mask, **kwargs)

    def __getstate__(self):
        ma = self.view(np.ma.MaskedArray).__getstate__()
        opt = self._optinfo
        return ma, opt

    def __setstate__(self, state):
        ma, opt = state
        super().__setstate__(ma)
        self._optinfo.update(opt)

    #        opt = self._optinfo
    #        return ma, opt
    #         # ma = self.view(np.ma.MaskedArray)
    #         # info = self._optinfo
    #         # return ma, info
    # self._optinfo.update(opt)
    # ma, info = state
    #         # super(MaskedlnPi, self).__setstate__(ma)
    #         # self._optinfo.update(info)

    @classmethod
    def from_table(
        cls, path, lnz, state_kws=None, sep=r"\s+", names=None, csv_kws=None, **kwargs
    ):
        """
        Create lnPi object from text file table with columns [n_0,...,n_ndim, lnpi]

        Parameters
        ----------
        path : string like
            file object to be read
        lnz : array-like
            beta*(chemical potential) for each component
        state_kws : dict, optional
            define state variables, like volume, beta
        sep : string, optional
            separator for file read
        names : column names
        csv_kws : dict, optional
            optional arguments to `pandas.read_csv`
        kwargs  : extra arguments
            Passed to lnPi constructor
        """
        lnz = np.atleast_1d(lnz)
        ndim = len(lnz)

        if names is None:
            names = ["n_{}".format(i) for i in range(ndim)] + ["lnpi"]

        if csv_kws is None:
            csv_kws = {}

        da = (
            pd.read_csv(path, sep=sep, names=names, **csv_kws)
            .set_index(names[:-1])["lnpi"]
            .to_xarray()
        )
        return cls(
            data=da.values,
            mask=da.isnull().values,
            lnz=lnz,
            state_kws=state_kws,
            **kwargs
        )

    @classmethod
    def from_dataarray(cls, da, state_as_attrs=None, **kwargs):
        """
        create a lnPi object from xarray.DataArray
        """

        kws = {}
        kws["data"] = da.values
        if "mask" in da.coords:
            kws["mask"] = da.mask.values
        else:
            kws["mask"] = da.isnull().values

        # where are state variables
        if state_as_attrs is None:
            state_as_attrs = bool(da.attrs.get("state_as_attrs", False))
        if state_as_attrs:
            # state variables from attrs
            c = da.attrs
        else:
            c = da.coords

        lnz = []
        state_kws = {}
        for k in da.attrs["dims_state"]:
            val = np.array(c[k])
            if "lnz" in k:
                lnz.append(val)
            else:
                state_kws[k] = val * 1
        kws["lnz"] = lnz
        kws["state_kws"] = state_kws

        # any overrides
        kwargs = dict(kws, **kwargs)
        return cls(**kwargs)

    def list_from_masks(self, masks, convention="image"):
        """
        create list of lnpis corresponding to masks[i]

        Parameters
        ----------
        masks : list
            masks[i] is the mask for i'th lnpi
        convention : str or bool
            convention of input masks
        Returns
        -------
        lnpis : list
            list of lnpis corresponding to each mask
        """

        return [
            self.or_mask(m) for m in masks_change_convention(masks, convention, False)
        ]

    def list_from_labels(
        self,
        labels,
        features=None,
        include_boundary=False,
        check_features=True,
        **kwargs
    ):
        """
        create list of lnpis corresponding to labels
        """

        masks, features = labels_to_masks(
            labels=labels,
            features=features,
            include_boundary=include_boundary,
            convention=False,
            check_features=check_features,
            **kwargs
        )
        return self.list_from_masks(masks, convention=False)
